
lib = require("stuff/lazy")()
local la = lib.stuff.la
lib.stuff.globalGuard.start()
getmetatable(_G).__index = function(self, k)
	assert(false, "Attempt to use undefined global: " .. k)
end

local Class = lib.stuff.Class

local bit = require("bit")

local function pointToInt(p)
	--return p[1] + 2^13 + 2^15 * (p[2] + 2^13)
	local off = bit.lshift(1, 13)
	return bit.bor(
		p[1] + off,
		bit.lshift(
			p[2] + off,
			15
		)
	)
end

local function intToPoint(p)
	local off = bit.lshift(1, 13)
	local mask = bit.lshift(1, 15)-1
	return la.vec2(
		bit.band(mask, p) - off,
		bit.rshift(p, 15) - off
	)
end


local dirs = {
	la.vec2(1,0),
	la.vec2(0,1),
	la.vec2(-1,0),
	la.vec2(0,-1),
}

local function getDir(i)
	return dirs[i % 4 + 1]
end

local function genBoard()

	local probChangeDir = 0.6
	local probWall = 0.3
	local probClone = 0.5
	local steps = 8
	
	local walls = {}
	local visited = {[pointToInt{0,0}]=true}
	
	local path = {}
	local clonePoints = {}
	local dir = math.floor(math.random()*4)

	local clones = {
		{
			pos = la.vec2(0,0),
			t = 1,
		},
	}

	local function getNextTile(clone)
		return clone.pos + getDir(path[clone.t])
	end
	
	for i = 1, steps do
		--[[
		What to do?
			- Randomly modify dir
			- Randomly spawn a new clone
			- For each clone
				- Randomly place wall on next tile (if not visited)
			- For each clone
				- Walk one step
		]]
		
		
		if math.random() < probChangeDir then
			dir = (dir + (math.random() < 0.5 and -1 or 1)) % 4
		end
		table.insert(path, dir)

		if #path > 2 and not clonePoints[#path] and math.random() < probClone then
			table.insert(
				clones,
				{
					pos = clones[1].pos:copy(),
					t = 1,
				}
			)
			table.insert(clonePoints, true)
		else
			table.insert(clonePoints, false)
		end

		-- For each clone place a wall in front if the next place is unvisited
		for _, c in ipairs(clones) do
			local next = pointToInt(getNextTile(c))
			if (not visited[next]) and math.random() < probWall then
				walls[next] = true
			end
		end

		-- Advance each clone one step (or bump into a wall)
		-- and mark new position as visited
		for _, c in ipairs(clones) do
			local next = getNextTile(c)
			if not walls[pointToInt(next)] then
				c.pos = next
			end
			visited[pointToInt(c.pos)] = true
			c.t = c.t + 1
		end
	end

	local targets = {}
	for _, c in ipairs(clones) do
		local p = pointToInt(c.pos)
		targets[p] = (targets[p] or 0) + 1
	end

	for p, _ in pairs(walls) do
		print("wall", intToPoint(p))
	end

	print("Musterlösung")
	for i, action in ipairs(path) do
		if action then
			print(clonePoints[i] and "clone" or "", ({
				"rechts",
				"runter",
				"links",
				"hoch",
			})[action+1])
			-- print(getDir(action))
		end
	end
	print()

	return {
		walls = walls,
		targets = targets,
		path = path,
	}
end

if false then
	for _, p in ipairs({
		{0,0},
		{-10,45},
		{100,50},
		{-331,-1652},
	}) do
		setmetatable(p, la.vec2)
		print("Point (before)", p)
		local i = pointToInt(p)
		print("Int", i)
		p = intToPoint(i)
		print("Point (after)", p)
	end
end


--[[
Spielfeld:
	- Spawn immer bei 0,0
	- Hindernisse sind table mit pointToInt als index und value immer true
		z.B. {[pointToInt{4,6}] = true}
	- Ziele wie Hindernisse

]]

local function getBoardBounds(points)
	local bounds = {la.vec2(0,0), la.vec2(0,0)}
	local funs = {math.min, math.max}

	for _, pos in ipairs(points) do
		for f = 1, 2 do
			for i = 1, 2 do
				bounds[f][i] = funs[f](
					bounds[f][i],
					pos[i]
				)
			end
		end
	end
	
	return bounds
end

local function drawShape(tbl, shapeFunc, fill, offset,  ...)
	for p, _ in pairs(tbl) do
		local x, y = unpack(intToPoint(p) + la.vec2(offset, offset))
		shapeFunc(fill, x, y, ...)
	end
end

local function drawBoard(targetSize, walls, targets)
	local bounds = getBoardBounds(walls, targets)
	bounds[1]:anadd(-1)
	bounds[2]:anadd(2)
	local boardSize = bounds[2] - bounds[1]

	local ratio = {}
	for i = 1, 2 do
		ratio[i] = targetSize[i] / boardSize[i]
	end
	local limiting = ratio[1] < ratio[2] and 1 or 2
	local other = 3 - limiting

	local scale = ratio[limiting]
	local offset = la.vec2(0,0)
	offset[other] = (targetSize[other] - boardSize[other]*scale) / 2

	love.graphics.push()
	love.graphics.translate(unpack(offset))
	love.graphics.scale(scale)
	love.graphics.translate(unpack(bounds[1]:nmul(-1)))
	
	love.graphics.setColor({1,1,1,1})

	love.graphics.setLineWidth(0.1)
	love.graphics.setColor({1,0,1,1})
	love.graphics.circle('line', 0.5, 0.5, 0.45)

	love.graphics.setColor({0.2,0.2,0.2,1})

	for x = bounds[1][1], bounds[2][1]-1, 1 do
	for y = bounds[1][2], bounds[2][2]-1, 1 do
		love.graphics.rectangle("fill", x+0.4, y+0.4, 0.2, 0.2)
	end end

	love.graphics.setColor({1,1,1,1})
	


	drawShape(walls, love.graphics.rectangle, "fill", 0.1, 0.8, 0.8)
	drawShape(targets, love.graphics.circle, "line", 0.5, 0.4)

	-- for p, _ in ipairs(walls) do
	-- 	local x, y = unpack(intToPoint(p) + la.vec2(0.1, 0.1))
	-- 	love.graphics.rectangle('fill', x, y, 0.8, 0.8)
	-- end
	-- for p, _ in ipairs(targets) do
	-- 	local x, y = unpack(intToPoint(p) + la.vec2(0.1, 0.1))
	-- 	love.graphics.circle('fill', x, y, 0.4)
	-- end
	love.graphics.pop()
end

local function parseBoard(map)
	local wallsAbs = {}
	local targetsAbs = {}
	local spawn

	local pos = la.vec2(0,0)
	for char in map:gmatch(".") do
		if char == "\n" then
			pos[2] = pos[2] + 1
			pos[1] = 0
		else
			if char == "#" then
				table.insert(wallsAbs, pos:copy())
			elseif char == "S" then
				spawn = pos:copy()
			elseif char == "o" then
				table.insert(targetsAbs, pos:copy())
			end
			pos[1] = pos[1] + 1
		end
	end
	
	local walls = {}
	local targets = {}

	for source, sink in pairs({
		[wallsAbs] = walls,
		[targetsAbs] = targets,
	}) do
		for i = 1, #source do
			local p = source[i] - spawn
			sink[pointToInt(p)] = true
		end
	end

	return walls, targets
end

local function drawBoardAnimated(
	targetSize,
	t,
	walls,
	targets,
	poss,
	actions
)


	local cloneInfos = {}

	for i, p in pairs(poss) do
		local ip = pointToInt(p)


		local action = actions[i]

		local scale = la.vec2(1,1)
		local offset = la.vec2(0,0)

		if action then
			local dir = getDir(action)
			local target = p + dir
			local a
			local squashAxis = dir[1] ~= 1 and 2 or 1
			local squash = math.cos(t * math.pi * 2) * 0.1
			if walls[pointToInt(target)] then
				a = (0.5 - math.cos(t * math.pi * 2) * 0.5) * 0.5
			else
				squashAxis = 3 - squashAxis
				a = 0.5 - 0.5 * math.cos(t * math.pi)
			end
			scale[squashAxis] = 0.9 + squash
			-- scale[3-squashAxis] = 1.1 - squash
			scale[3-squashAxis] = 1.0 / scale[squashAxis]
			offset = dir:nmul(a)
		end

		local b = i == 1 and 1.0 or 0.6

		table.insert(
			cloneInfos,
			{
				color = {b,b,b,1},
				corner = offset + p,
				scale = scale,
			}
		)

	end

	print(#cloneInfos)

	local circleSegments = 48
	local points = {}
	for _, tbl in ipairs({walls, targets}) do
		for p, _ in pairs(tbl) do
			table.insert(points, intToPoint(p))
		end
	end
	for _, clone in ipairs(cloneInfos) do
		table.insert(points, clone.corner)
	end

	local bounds = getBoardBounds(points)
	bounds[1]:anadd(-1)
	bounds[2]:anadd(2)
	local boardSize = bounds[2] - bounds[1]

	local ratio = {}
	for i = 1, 2 do
		ratio[i] = targetSize[i] / boardSize[i]
	end
	local limiting = ratio[1] < ratio[2] and 1 or 2
	local other = 3 - limiting

	local scale = ratio[limiting]
	local offset = la.vec2(0,0)
	offset[other] = (targetSize[other] - boardSize[other]*scale) / 2

	love.graphics.push()
	love.graphics.translate(unpack(offset))
	love.graphics.scale(scale)
	love.graphics.translate(unpack(bounds[1]:nmul(-1)))

	love.graphics.setColor({0.2,0.2,0.2,1})
	for x = math.floor(bounds[1][1]), math.floor(bounds[2][1]-1), 1 do
	for y = math.floor(bounds[1][2]), math.floor(bounds[2][2]-1), 1 do
		love.graphics.rectangle("fill", x+0.4, y+0.4, 0.2, 0.2)
	end end

	love.graphics.setColor({.9,.9,.9,1})
	for p in pairs(walls) do
		local p = intToPoint(p)
		love.graphics.rectangle("fill", p[1]+0.2, p[2]+0.2, 0.6, 0.6)
	end

	love.graphics.setLineWidth(0.075)
	love.graphics.setColor({1,0,1,1})
	for p in pairs(targets) do
		local p = intToPoint(p)
		love.graphics.circle("line", p[1]+0.5, p[2]+0.5, 0.4, circleSegments)
	end

	
	for _, clone in ipairs(cloneInfos) do
		love.graphics.setColor(clone.color)
		love.graphics.push()
		love.graphics.translate(unpack(clone.corner + la.vec2(.5, .5)))
		love.graphics.scale(unpack(clone.scale))
		love.graphics.circle("fill", 0, 0, 0.35, circleSegments)
		love.graphics.pop()
	end

	love.graphics.setColor({1,1,1,1})
	love.graphics.setLineWidth(0.075)
	love.graphics.circle("line", 0.5, 0.5, 0.3)

	love.graphics.pop()
end




------------------------------------
---------- Gamestate stuff starts here
------------------------------------

local GameState = lib.stuff.Class("GameState")

function GameState:__init(walls, targets)
	self.walls = walls
	self.targets = targets

	self.clonePos = {la.vec2(0,0)}
	self.cloneT = {1}
end

function GameState:getNumClones()
	return #self.clonePos
end

function GameState:getNumSolved()
	local numTotal = 0
	local numUnsolved = 0
	local unsolved = {}
	for k, v in pairs(self.targets) do
		unsolved[k] = v
		numTotal = numTotal + 1
	end
	for _, p in pairs(self.clonePos) do
		unsolved[pointToInt(p)] = nil
	end
	for k, v in pairs(unsolved) do
		numUnsolved = numUnsolved + 1
	end
	return numTotal - numUnsolved, numTotal
end

function GameState:getNextPos(path, cloneNum)
	local p = self.clonePos[cloneNum]
	local dir = getDir(
		path[self.cloneT[cloneNum]]
	)
	local target = p + dir
	if self.walls[pointToInt(target)] then
		return p:copy()
	else
		return target
	end
end

function GameState:cloneStep(clonePoints)
	local nextState = GameState(self.walls, self.targets)
	local t = self.cloneT[1]
	
	-- Copy all clones into next state
	for i, p in ipairs(self.clonePos) do
		nextState.clonePos[i] = p:copy()
		nextState.cloneT[i] = self.cloneT[i]
	end

	-- Create clone in next state (if necessary)
	if clonePoints[t] then
		local new = #self.clonePos + 1
		nextState.clonePos[new] = self.clonePos[1]:copy()
		nextState.cloneT[new] = 1
	end

	return nextState
end

function GameState:moveStep(path)
	local nextState = GameState(self.walls, self.targets)
	
	for i, _ in ipairs(self.clonePos) do
		local next = self:getNextPos(path, i)
		nextState.clonePos[i] = next
		nextState.cloneT[i] = self.cloneT[i] + 1
	end

	return nextState
end

function GameState:getDrawParams(path, clonePoints)
	local actions = {}
	for i, t in ipairs(self.cloneT) do
		actions[i] = path[t] or false
	end
	return self.walls, self.targets, self.clonePos, actions
end


------------------------------------
---------- Game starts here
------------------------------------

local Game = lib.stuff.Class("Game")

Game.font = love.graphics.newFont(24)

function Game:__init(walls, targets, numLevel, maxSteps)
	self.numLevel = numLevel
	self.maxSteps = maxSteps
	
	self.walls = walls
	self.targets = targets

	self.state = GameState(walls, targets)
	self.path = {}
	self.clonePoints = {}

	self.animAction = 0
	self.animProgress = 0

	self.isSolved = false
end

function Game:playerClone()
	self.clonePoints[#self.path+1] = true
end

function Game:playerMove(dir)
	table.insert(self.path, dir)
	local i = #self.path
	self.clonePoints[i] = self.clonePoints[i] or false
end

function Game:playerUndo()
	local n = #self.path
	if self.clonePoints[n+1] then
		self.clonePoints[n+1] = nil
	else
		self.clonePoints[n+1] = nil
		self.clonePoints[n] = nil
		self.path[n] = nil
	end
end

function Game:playerReset()
	self.animProgress = 0
	self.animAction = 0
	self.path = {}
	self.clonePoints = {}
end

function Game:update(dt)
	self.animProgress = self.animProgress + dt*8 --*math.max(4, #self.path - self.animAction)

	while self.animProgress > 1 do
		self.animProgress = self.animProgress - 1
		self.animAction = self.animAction + 1
	end

	if self.animAction >= #self.path then
		self.animProgress = 0
		self.animAction = #self.path
	end
end

function Game:getState()
	local state = GameState(self.walls, self.targets)
	for i = 1, #self.path do
		if i > self.animAction then
			break
		end
		state = state:moveStep(self.path)
		state = state:cloneStep(self.clonePoints)
	end
	return state
end

function Game:getStats()
	local state = self:getState()
	local solved, required = state:getNumSolved()
	return {
		solved = solved,
		required = required,
		clones = state:getNumClones(),
		steps = #self.path,
		level = self.numLevel,
	}
	-- Irgendwie anzeigen
	-- solved/required solved
	-- numClones/required used

end

function Game:draw()

	local state = self:getState()

	drawBoardAnimated(
		la.vec2(love.graphics.getDimensions()),
		--lastClone and 0 or self.animProgress,
		self.animProgress,
		state:getDrawParams(self.path, self.clonePoints)
	)

	love.graphics.setColor({1,1,1,0.8})

	local white = {1,1,1,0.8}
	local pink = {1,0,1,0.8}
	local red = {1, 0.5, 0.5, 0.9}

	local stats = self:getStats()

	local text = {}

	self.isSolved = stats.solved == stats.required and (
		(not self.maxSteps) or stats.steps <= self.maxSteps
	)


	-- white,
	-- tostring(stats.solved) .. "/" .. tostring(stats.required) .. " Done" .. "\n"

	local function pt(t, color)
		table.insert(text, color or white)
		table.insert(text, t)
	end

	pt(tostring(stats.solved))
	pt("/")
	pt(tostring(stats.required))
	pt(" Done\n")
	local sc = (not self.maxSteps or stats.steps <= self.maxSteps) and white or red
	pt(tostring(stats.steps), sc)
	if self.maxSteps then
		pt("/", sc)
		pt(tostring(self.maxSteps), sc)
	end
	pt(" Steps\n", sc)


	if self.isSolved then
		pt("\n\nSolved!", pink)
		pt("\nPress ")
		pt("return", pink)
		pt(" for next level")
	end

	love.graphics.printf(text, self.font, 20, 20, 1000)

	-- love.graphics.printf(stats.solved, self.font, 20, 20, 1000,)
	-- love.graphics.printf(stats.required, self.font, 20, 40, 1000,)
	-- love.graphics.printf(stats.numClones, self.font, 20, 60, 1000,)
	

end


local game
local numLevel = 0

local function nextLevel()
	numLevel = numLevel + 1
	local level = lib.levels[numLevel]
	local w, t = parseBoard(level.map) 
	game = Game(w, t, numLevel, level.steps)
end

function love.load()
	nextLevel()
end

function love.update(dt)
	game:update(dt)
end

function love.draw()
	game:draw()
end

function love.keypressed(k)
	local delta = ({
		right = 0,
		down = 1,
		left = 2,
		up = 3,
	})[k]

	if delta then
		game:playerMove(delta)
	end

	if k == "space" then
		game:playerClone()
	end
	
	if k == "backspace" then
		game:playerUndo()
	end

	if k == "r" then
		game:playerReset()
	end

	if k == "return" then
		if game.isSolved then
			nextLevel()
		end
	end
end

if true then
	return
end

------------------------------------
---------- Runtime stuff starts here!
------------------------------------


local board
local gameState
local seed = 0
local path
local clonePoints

local function makeBoard(delta)
	seed = seed + (delta or 1)
	math.randomseed(seed)
	board = genBoard()
	gameState = GameState(board.walls, board.targets)
	path = {}
	clonePoints = {}
end

function love.load()
	love.keyboard.setKeyRepeat(true)
	makeBoard()
end

function love.draw()
	local ps = {
		{0,0},
		{1,1},
		{2,2},
		{3,3},
		{4,4},
	}

	local ps2 = {}
	for _, p in ipairs(ps) do
		ps2[pointToInt(p)] = 1
	end

	--[[
	drawBoard(
		la.vec2(love.graphics.getDimensions()),
		board.walls,
		board.targets
	)
	]]

	local t = ((love.timer.getTime()*2) % 3) - 1
	t = math.max(0, math.min(1, t))
	t = 0

	drawBoardAnimated(
		la.vec2(love.graphics.getDimensions()),
		t,
		gameState:getDrawParams(path, clonePoints)
	)

	love.graphics.setColor{1,1,1,1}
	love.graphics.scale(3)
	love.graphics.print(tostring(seed))
	--love.graphics.rectangle("fill", 0, 0, 10, 10)
end

function love.keypressed(k)
	local delta = ({
		a = -1,
		d = 1,
	})[k]

	if delta then
		makeBoard(delta)
	end

	delta = ({
		right = 0,
		down = 1,
		left = 2,
		up = 3,
	})[k]

	if delta then
		table.insert(path, delta)
		gameState = gameState:moveStep(path)
	end

	if love.keyboard.isDown("space") then
		if not clonePoints[#path+1] then
			clonePoints[#path+1] = true
			gameState = gameState:cloneStep(clonePoints)
		end
	end
end
