-- Stuff to do with iteration

local it = {}

-- Creating iterators

function it.range(a, b, c)
	c = c or 1
	return function (s, i)
		i = i + s[2]
		if s[1] < i then
			return nil
		end
		return i, i
	end, {b, c}, a - c
end

-- Modifying iterators

function it.map(fun, f, s, var)
	return function()
		local args = {f(s, var)}
		if args[1] == nil then
			return nil
		end
		var = args[1]
		return fun(unpack(args))
	end
end

function it.zip(...)
	local iters = {...}
	return it.listComp(
		function(_, iter)
			local args = {iter[1](iter[2], iter[3])}
			iter[3] = args[1]
			return args
		end,
		ipairs(iters)
	)
end

-- Evaluating iterators

function it.ffor(fun, f, s, var)
	while true do
		local args = {f(s, var)}
		if args[1] == nil then
			break
		end
		var = args[1]
		fun(unpack(args))
	end
end

function it.listify(...)
	local accum = {}
	for _, i in ... do
		table.insert(accum, i)
	end
	return accum
end

function it.dictify(...)
	local accum = {}
	for k, v in ... do
		accum[k] = v
	end
	return accum
end

-- Shortcuts

function it.listComp(fun, ...)
	return it.listify(
		it.map(
			fun,
			...
		)
	)
end

function it.dictComp(fun, ...)
	return it.dictify(
		it.map(
			fun,
			...
		)
	)
end

return it
